/*
 * Copyright (c) 2012-2013 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 *					Anthony Mallet on Thu Aug  2 2012
 */
#include <sys/types.h>
#include <assert.h>
#include <err.h>
#include <fcntl.h>
#include <getopt.h>
#include <pthread.h>
#include <unistd.h>

#include <cerrno>

#include "ros/init.h"
#include "ros/master.h"
#include "ros/callback_queue.h"
#include "ros/callback_queue_interface.h"

#include "genom3/c/client.h"
#include "private.h"


/* --- local data ---------------------------------------------------------- */

#define ros_cbevent	'e'
#define ros_end		'q'

static int ros_nodeinit = 0;
static int ros_evpipe[2] = { -1, -1 };
static pthread_t ros_evtid;
static int ros_evtid_init = 0;

static void *	ros_evtask(void *data);

genom_event genom_weak genom_genom_ros_id;
struct genom_genom_ros_detail { char what[256]; };

static inline genom_event
make_genom_syserr(int code) {
  genom_syserr_detail d = { code };
  return genom_syserr(&d);
}


/* --- ros_init ------------------------------------------------------------ */

genom_event
ros_init(int argc, char *argv[])
{
  int s;

  /* global ros init */
  if (!ros_nodeinit) {
    try {
      ros::init(argc, argv, "genom_c_client",
                ros::init_options::NoSigintHandler |
                ros::init_options::AnonymousName |
                ros::init_options::NoRosout);
      if (!ros::master::check()) {
        genom_genom_ros_detail d;
        snprintf(d.what, sizeof(d.what), "roscore may not be running on %s:%u",
                 ros::master::getHost().c_str(), ros::master::getPort());
        return genom_throw(genom_genom_ros_id, &d, sizeof(d));
      }
      ros::start();
    }
    catch (ros::Exception e) {
      genom_genom_ros_detail d;
      snprintf(d.what, sizeof(d.what), "%s", e.what());
      return genom_throw(genom_genom_ros_id, &d, sizeof(d));
    }

    if (pipe(ros_evpipe) ||
        fcntl(ros_evpipe[0], F_SETFL, O_NONBLOCK))
      goto error;
  }
  ros_nodeinit++;

  /* create event task */
  if (!ros_evtid_init) {
    s = pthread_create(&ros_evtid, NULL, ros_evtask, NULL);
    if (s) goto error;
    ros_evtid_init = 1;
  }

  return 0;

error:
  return make_genom_syserr(errno);
}


/* --- ros_fini ------------------------------------------------------------ */

void
ros_fini(void)
{
  int i;

  if (ros_nodeinit) ros_nodeinit--;
  if (ros_nodeinit) return;

  ros::shutdown();

  if (ros_evtid_init) {
    pthread_join(ros_evtid, NULL);
    ros_evtid_init = 0;
  }

  for(i=0; i<2; i++) {
    if (ros_evpipe[i] > 0) {
      close(ros_evpipe[i]);
      ros_evpipe[i] = -1;
    }
  }
}


/* --- ros_getopt ---------------------------------------------------------- */

/* options */
static const char shortopts_string[] = "+i:";
static struct option longopts_list[] = {
  { "name",		required_argument,	NULL,	'i'},
  { NULL, 0, NULL, 0}
};

genom_event
ros_getopt(int argc, char *argv[], struct ros_options *opts)
{
  int c;

  opts->instance = NULL;

  /* reset getopt */
#ifdef HAVE_OPTRESET
  optreset = 1;
#endif
#ifdef __GLIBC__
  optind = 0; /* not "1", see gnu manual for getopt(3) ... */
#else
  optind = 1;
#endif

  /* parse command line options */
  opterr = 0;
  while (
    (c = getopt_long(argc, argv, shortopts_string, longopts_list, NULL)) != -1)
    switch (c) {
      case 0: break;

      case 'i': opts->instance = optarg; break;

      case '?':
      default:
        genom_genom_ros_detail d;
        snprintf(d.what, sizeof(d.what), "unknown option %s", argv[optind]);
        return genom_throw(genom_genom_ros_id, &d, sizeof(d));
    }

  return 0;
}


/* --- ros_newrqst --------------------------------------------------------- */

genom_event
ros_newrqst(genom_client h, size_t outsize, int *rqstid)
{
  genom_client_s::genom_client_request *a;
  int i;
  assert(h);

  *rqstid = -1;
  pthread_mutex_lock(&h->activelk);
  for(i = h->next; i < h->next + h->nactive; i++)
    if (h->active_rqst[i%h->nactive].status == ROS_CLIENT_FREE) {
      *rqstid = i%h->nactive;
      break;
    }
  if (*rqstid < 0) {
    a = (genom_client_s::genom_client_request *)
      realloc(h->active_rqst, (h->nactive+1)*sizeof(*a));
    if (!a) return make_genom_syserr(errno);

    *rqstid = h->nactive;
    h->nactive++;
    h->active_rqst = a;
  }
  h->next = *rqstid + 1;

  h->active_rqst[*rqstid].info = NULL;
  h->active_rqst[*rqstid].aid = -1;
  h->active_rqst[*rqstid].status = ROS_CLIENT_WAIT_ACK;
  h->active_rqst[*rqstid].pstatus = ROS_CLIENT_WAIT_ACK;
  h->active_rqst[*rqstid].error = 0;
  h->active_rqst[*rqstid].exdetail = NULL;
  h->active_rqst[*rqstid].output = malloc(outsize);
  h->active_rqst[*rqstid].sentcb = NULL;
  h->active_rqst[*rqstid].donecb = NULL;
  h->active_rqst[*rqstid].cb_data = NULL;
  if (!h->active_rqst[*rqstid].output) {
    h->active_rqst[*rqstid].status = ROS_CLIENT_FREE;
    pthread_mutex_unlock(&h->activelk);
    return make_genom_syserr(errno);
  }
  pthread_mutex_unlock(&h->activelk);

  return genom_ok;
}


/* --- ros_eventfd --------------------------------------------------------- */

int
ros_eventfd(void)
{
  return ros_evpipe[0];
}


/* --- ros_service_info ---------------------------------------------------- */

const struct genom_service_info *
ros_service_info(genom_client h, int rqstid)
{
  const struct genom_service_info *info = NULL;

  pthread_mutex_lock(&h->activelk);
  if (rqstid >= 0 && rqstid < h->nactive)
    if (h->active_rqst[rqstid].status != ROS_CLIENT_FREE)
      info = h->active_rqst[rqstid].info;
  pthread_mutex_unlock(&h->activelk);

  return info;
}


/* --- ros_waitevent ------------------------------------------------------- */

genom_event
ros_waitevent(void)
{
  fd_set fdset;
  int s;

  FD_ZERO(&fdset);
  do {
    FD_SET(ros_evpipe[0], &fdset);
    s = select(ros_evpipe[0]+1, &fdset, NULL, NULL, NULL);
  } while(s < 0 && errno == EINTR);

  if (s < 0) return make_genom_syserr(errno);
  return genom_ok;
}


/* --- ros_waitack --------------------------------------------------------- */

genom_event
ros_waitack(genom_client h, int rqstid)
{
  genom_event s;

  pthread_mutex_lock(&h->activelk);
  if (rqstid < 0 || rqstid >= h->nactive) {
    pthread_mutex_unlock(&h->activelk);
    return make_genom_syserr(ENOENT);
  }

  while(h->active_rqst[rqstid].pstatus == ROS_CLIENT_WAIT_ACK) {
    pthread_mutex_unlock(&h->activelk);
    s = ros_waitevent();
    if (s) return s;

    s = ros_doevents(h);
    if (s) return s;

    pthread_mutex_lock(&h->activelk);
  }

  pthread_mutex_unlock(&h->activelk);
  return genom_ok;
}


/* --- ros_waitterm -------------------------------------------------------- */

genom_event
ros_waitterm(genom_client h, int rqstid)
{
  genom_event s;

  pthread_mutex_lock(&h->activelk);
  if (rqstid < 0 || rqstid >= h->nactive) {
    pthread_mutex_unlock(&h->activelk);
    return make_genom_syserr(ENOENT);
  }

  while(h->active_rqst[rqstid].pstatus == ROS_CLIENT_WAIT_ACK ||
        h->active_rqst[rqstid].status == ROS_CLIENT_WAIT_TERM) {
    pthread_mutex_unlock(&h->activelk);
    s = ros_waitevent();
    if (s) return s;

    s = ros_doevents(h);
    if (s) return s;

    pthread_mutex_lock(&h->activelk);
  }

  pthread_mutex_unlock(&h->activelk);
  return genom_ok;
}


/* --- ros_done ------------------------------------------------------------ */

int
ros_done(genom_client h, int rqstid)
{
  int done = 1;

  pthread_mutex_lock(&h->activelk);
  if (rqstid >= 0 || rqstid < h->nactive)
    if (h->active_rqst[rqstid].status == ROS_CLIENT_WAIT_ACK ||
        h->active_rqst[rqstid].status == ROS_CLIENT_WAIT_TERM)
      done = 0;
  pthread_mutex_unlock(&h->activelk);

  return done;
}


/* --- ros_clean ----------------------------------------------------------- */

genom_event
ros_clean(genom_client h, int rqstid)
{
  struct genom_client_s::genom_client_request *a;

  pthread_mutex_lock(&h->activelk);
  if (rqstid < 0 || rqstid >= h->nactive) {
    pthread_mutex_unlock(&h->activelk);
    return make_genom_syserr(ENOENT);
  }

  a = &h->active_rqst[rqstid];
  if (a->status == ROS_CLIENT_FREE) {
    pthread_mutex_unlock(&h->activelk);
    return make_genom_syserr(ENOENT);
  }

  if (a->status == ROS_CLIENT_WAIT_ACK ||
      a->status == ROS_CLIENT_WAIT_TERM) {
    // XXX how to cancel on the server?
  }

  a->status = ROS_CLIENT_FREE;
  if (a->output) {
    if (a->info->fini_output) a->info->fini_output(a->output);
    free(a->output);
  }
  if (a->exdetail) {
    /* XXX do not leak dynamically allocated members */
    free(const_cast<void *>(a->exdetail));
  }
  pthread_mutex_unlock(&h->activelk);
  return genom_ok;
}


/* --- ros_doevents -------------------------------------------------------- */

genom_event
ros_doevents(genom_client h)
{
  struct genom_client_s::genom_client_request *a;
  ssize_t r;
  char e[16];
  int i;

  if (ros_evpipe[0] < 0)
    return make_genom_syserr(EBADF);

  do {
    do {
      r = read(ros_evpipe[0], e, sizeof(e));
    } while (r == -1 && errno == EINTR);
    if (r == 0) return make_genom_syserr(EBADF);

    if (r == -1) {
      if ((errno == EAGAIN || errno == EWOULDBLOCK)) break;
      return make_genom_syserr(errno);
    }

    for(i=0; i<r; i++) {
      switch(e[i]) {
        case ros_end:
          return make_genom_syserr(ENODEV);

        case ros_cbevent:
          break;

        default: assert(!"reached unreachable");
      }
    }
  } while(r == sizeof(e));

  pthread_mutex_lock(&h->activelk);
  if (!h->events) {
    pthread_mutex_unlock(&h->activelk);
    return genom_ok;
  }
  h->events = 0;
  for(i = 0; i < h->nactive; i++) {
    a = &h->active_rqst[i];
    if (a->status == a->pstatus) continue;

    switch(a->status) {
      case ROS_CLIENT_WAIT_TERM:
        a->pstatus = a->status;
        if (a->sentcb)
          (*a->sentcb)(h, i, 0, a->error, NULL, a->exdetail, a->cb_data);
        break;

      case ROS_CLIENT_TERM:
        a->pstatus = a->status;
        if (a->donecb) {
          if (a->error)
            (*a->donecb)(h, i, 1, a->error, NULL, a->exdetail, a->cb_data);
          else
            (*a->donecb)(h, i, 1, genom_ok, a->output, NULL, a->cb_data);
        }
        break;
    }
  }
  pthread_mutex_unlock(&h->activelk);

  return genom_ok;
}


/* --- fake callback for services ------------------------------------------ */

static struct ros_fake_cb_ : public ros::CallbackInterface {
  CallResult call(void) { return Success; }
} ros_fake_cb;

struct ros_fake_delete {
  void operator()(ros_fake_cb_ *) {}
};

void
ros_evtask_signal(void)
{
  ros::CallbackInterfacePtr cb(&ros_fake_cb, ros_fake_delete());
  ros::getGlobalCallbackQueue()->addCallback(cb);
}


/* --- ros_evtask ---------------------------------------------------------- */

static void *
ros_evtask(void *data)
{
  ros::CallbackQueue *queue = ros::getGlobalCallbackQueue();
  char c;

  c = ros_cbevent;
  while(ros::ok()) {
    queue->callAvailable(ros::WallDuration(10.));
    if (!ros_nodeinit) break;
    write(ros_evpipe[1], &c, 1);
  }

  c = ros_end;
  write(ros_evpipe[1], &c, 1);
  return NULL;
}
